var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
import CodeMirror from 'codemirror';
import 'codemirror/mode/gfm/gfm.js';
import 'codemirror/addon/display/placeholder.js';
import _ from 'lodash-es';
import prettier from 'prettier/standalone';
import parserMarkdown from 'prettier/parser-markdown';
import { DEFAULT_OPTIONS, DEFAULT_FROM_TEXTAREA_OPTIONS, } from './markdown-editor-options';
var MarkdownEditorBase = /** @class */ (function () {
    function MarkdownEditorBase(codemirror, options) {
        this.cm = codemirror;
        this.options = options;
        this.applyCodemirrorOptions();
        this.applyEditorKeyMappings();
        this.removeLinkClassFromImageTexts();
    }
    /***** Basic Editor API *****/
    /**
     * Toggle "bold" for each selection.
     */
    MarkdownEditorBase.prototype.toggleBold = function () {
        var preferred = this.options.preferredTokens.bold;
        this.toggleInlineFormatting(preferred, MarkdownEditorBase.BOLD_TOKENS.filter(function (t) { return t !== preferred; }));
    };
    /**
     * Toggle "italic" for each selection.
     */
    MarkdownEditorBase.prototype.toggleItalic = function () {
        var preferred = this.options.preferredTokens.italic;
        this.toggleInlineFormatting(preferred, MarkdownEditorBase.ITALIC_TOKENS.filter(function (t) { return t !== preferred; }), true);
    };
    /**
     * Toggle "strikethrough" for each selection.
     */
    MarkdownEditorBase.prototype.toggleStrikethrough = function () {
        this.toggleInlineFormatting('~~');
    };
    /**
     * Toggle "inline code" for each selection.
     */
    MarkdownEditorBase.prototype.toggleInlineCode = function () {
        this.toggleInlineFormatting('`');
    };
    /**
     * Toggle inline formatting for each selection by wrapping it with the specified token
     * and unwrapping it with the specified token or one of the alternative tokens.
     * @param token the token
     * @param altTokens the alternative tokens
     */
    MarkdownEditorBase.prototype.toggleInlineFormatting = function (token, altTokens, preventDoubledTokenMatch) {
        if (altTokens === void 0) { altTokens = []; }
        if (preventDoubledTokenMatch === void 0) { preventDoubledTokenMatch = false; }
        var newSelections = [];
        var selections = _.cloneDeep(this.cm.listSelections());
        var _loop_1 = function (i) {
            var oldSelection = selections[i];
            var newSelection = _.cloneDeep(oldSelection);
            var from = oldSelection.from();
            var to = oldSelection.to();
            var endLineLength = this_1.cm.getLine(to.line).length;
            var linePartBefore = this_1.cm.getRange({ line: from.line, ch: 0 }, from);
            var linePartAfter = this_1.cm.getRange(to, { line: to.line, ch: endLineLength });
            var prefixToken = __spreadArrays([token], altTokens).find(function (t) {
                var escapedToken = escapeRegexChars(t);
                var lookBehind = preventDoubledTokenMatch ? '(?<!' + escapedToken + ')' : '';
                return linePartBefore.search(RegExp(lookBehind + escapedToken + '$')) > -1;
            });
            var suffixToken = __spreadArrays([token], altTokens).find(function (t) {
                var escapedToken = escapeRegexChars(t);
                var lookAhead = preventDoubledTokenMatch ? '(?!' + escapedToken + ')' : '';
                return linePartAfter.search(RegExp('^' + escapedToken + lookAhead)) > -1;
            });
            // indicate whether the tokens before/after the selection have been inserted or deleted
            var beforeShift = 0;
            var afterShift = 0;
            // Insert or delete tokens depending whether they exist
            if (suffixToken) {
                var suffixEnd = {
                    line: to.line,
                    ch: to.ch <= endLineLength - suffixToken.length ? to.ch + suffixToken.length : endLineLength,
                };
                this_1.cm.replaceRange('', to, suffixEnd, '+toggleBlock');
                afterShift = -1;
            }
            else {
                this_1.cm.replaceRange(token, to, undefined, '+toggleBlock');
                afterShift = 1;
            }
            if (prefixToken) {
                var prefixStart = { line: from.line, ch: from.ch >= prefixToken.length ? from.ch - prefixToken.length : 0 };
                this_1.cm.replaceRange('', prefixStart, from, '+toggleBlock');
                beforeShift = -1;
            }
            else {
                this_1.cm.replaceRange(token, from, undefined, '+toggleBlock');
                beforeShift = 1;
            }
            // Adjust selections to originally selected characters
            if (oldSelection.empty())
                newSelection.head = newSelection.anchor;
            else if (from.line === to.line)
                newSelection.to().ch += beforeShift * token.length;
            newSelection.from().ch += beforeShift * token.length;
            newSelections.push(newSelection);
            // Adjust all following selections to originally selected characters
            for (var j = i + 1; j < selections.length; j++) {
                var s = selections[j];
                if (s.empty()) {
                    s.head = s.anchor;
                }
                else {
                    if (s.head.line === from.line)
                        s.head.ch += beforeShift * token.length;
                    if (s.head.line === to.line)
                        s.head.ch += afterShift * token.length;
                }
                if (s.anchor.line === from.line)
                    s.anchor.ch += beforeShift * token.length;
                if (s.anchor.line === to.line)
                    s.anchor.ch += afterShift * token.length;
                else
                    break;
            }
        };
        var this_1 = this;
        for (var i = 0; i < selections.length; i++) {
            _loop_1(i);
        }
        this.cm.setSelections(newSelections, undefined, { origin: '+toggleBlock' });
        this.cm.focus();
    };
    /**
     * Set the specified heading level for each selected line. If `level` is 0, the heading token is removed.
     * @param level the heading level
     */
    MarkdownEditorBase.prototype.setHeadingLevel = function (level) {
        var headingToken = '#'.repeat(level) + (level === 0 ? '' : ' ');
        this.replaceTokenAtLineStart(function (oldLineContent) { return oldLineContent.replace(/^((#)*( )?)/, headingToken); });
    };
    /**
     * Increase the heading level for each selected line, i.e. make the heading smaller.
     * If the current heading level is at the maximum of 6, the heading token is just removed.
     * If the current heading level is 0 (no heading), the new heading level is set to 1.
     */
    MarkdownEditorBase.prototype.increaseHeadingLevel = function () {
        this.replaceTokenAtLineStart(function (oldLineContent) {
            var currentHeadingLevel = oldLineContent.search(/(?<=(^((#)+ ))).*/) - 1;
            if (currentHeadingLevel < 0)
                currentHeadingLevel = 0;
            var newLevel = currentHeadingLevel < 6 ? currentHeadingLevel + 1 : 0;
            var headingToken = '#'.repeat(newLevel) + (newLevel === 0 ? '' : ' ');
            return oldLineContent.replace(/^((#)*( )?)/, headingToken);
        });
    };
    /**
     * Decrease the heading level for each selected line, i.e. make the heading bigger.
     * If the current heading level is at the minimum of 1, the heading token is just removed.
     * If the current heading level is 0 (no heading), the new heading level is set to 6.
     */
    MarkdownEditorBase.prototype.decreaseHeadingLevel = function () {
        this.replaceTokenAtLineStart(function (oldLineContent) {
            var currentHeadingLevel = oldLineContent.search(/(?<=(^((#)+ ))).*/) - 1;
            var newLevel = currentHeadingLevel > 0 ? currentHeadingLevel - 1 : 6;
            var headingToken = '#'.repeat(newLevel) + (newLevel === 0 ? '' : ' ');
            return oldLineContent.replace(/^((#)*( )?)/, headingToken);
        });
    };
    /**
     * Toggle "quote" for each selected line.
     */
    MarkdownEditorBase.prototype.toggleQuote = function () {
        this.replaceTokenAtLineStart(function (oldLineContent) {
            // Has selected line a quote token?
            if (oldLineContent.search(/^>(\t| )*/) === -1) {
                return '> ' + oldLineContent;
            }
            else {
                return oldLineContent.replace(/^>(\t| )*/, '');
            }
        });
    };
    /**
     * Toggle "unordered list" for each selected line. Furthermore, a selected ordered list line is
     * transformed to an unordered list.
     */
    MarkdownEditorBase.prototype.toggleUnorderedList = function () {
        var preferred = this.options.preferredTokens.unorderedList + ' ';
        this.replaceTokenAtLineStart(function (oldLineContent) {
            // Has selected line a check list token?
            if (oldLineContent.search(MarkdownEditorBase.CHECK_LIST_PATTERN) !== -1) {
                return oldLineContent.replace(MarkdownEditorBase.CHECK_LIST_PATTERN, preferred);
            }
            // Has selected line an enumeration token?
            if (oldLineContent.search(MarkdownEditorBase.ORDERED_LIST_PATTERN) !== -1) {
                return oldLineContent.replace(MarkdownEditorBase.ORDERED_LIST_PATTERN, preferred);
            }
            // Has selected line a bullet point token?
            if (oldLineContent.search(MarkdownEditorBase.UNORDERED_LIST_PATTERN) !== -1) {
                return oldLineContent.replace(MarkdownEditorBase.UNORDERED_LIST_PATTERN, '');
            }
            else {
                return preferred + oldLineContent;
            }
        });
    };
    /**
     * Toggle "ordered list" for each selected line. Furthermore, a selected unordered list line is
     * transformed to an ordered list. Additionally adjusts the subsequent lines that are connected
     * to the list of the selected line.
     */
    MarkdownEditorBase.prototype.toggleOrderedList = function () {
        var _this = this;
        this.replaceTokenAtLineStart(function (oldLineContent, lineNumber) {
            // Has selected line an enumeration token?
            if (oldLineContent.search(MarkdownEditorBase.ORDERED_LIST_PATTERN) === -1) {
                var prevLine = _this.cm.getLine(lineNumber - 1);
                var listNumber = void 0;
                // Is previous line already enumerated list? -> Determine enumeration token for selected line
                if (!prevLine || prevLine.search(MarkdownEditorBase.ORDERED_LIST_PATTERN) === -1) {
                    listNumber = 1;
                }
                else {
                    var dotPos = prevLine.search(/\./);
                    listNumber = +prevLine.substring(0, dotPos) + 1;
                }
                _this.processNextLinesOfOrderedList(lineNumber, listNumber);
                var numberToken = listNumber + '. ';
                // Has selected line a check list token?
                if (oldLineContent.search(MarkdownEditorBase.CHECK_LIST_PATTERN) !== -1) {
                    return oldLineContent.replace(MarkdownEditorBase.CHECK_LIST_PATTERN, numberToken);
                }
                // Has selected line a bullet point token?
                if (oldLineContent.search(MarkdownEditorBase.UNORDERED_LIST_PATTERN) !== -1) {
                    return oldLineContent.replace(MarkdownEditorBase.UNORDERED_LIST_PATTERN, numberToken);
                }
                return numberToken + oldLineContent;
            }
            else {
                _this.processNextLinesOfOrderedList(lineNumber, 0);
                return oldLineContent.replace(MarkdownEditorBase.ORDERED_LIST_PATTERN, '');
            }
        });
    };
    /**
     * Adjust the enumeration of subsequent lines in same ordered list as line *baseLineNumber*.
     * @param baseLineNumber the selected line which is toggled
     * @param baseListNumber the list number of the selected line (should be 0, if list starts after selected line)
     */
    MarkdownEditorBase.prototype.processNextLinesOfOrderedList = function (baseLineNumber, baseListNumber) {
        var listNumber = baseListNumber;
        var nextLineNumber = baseLineNumber + 1;
        var nextLine = this.cm.getLine(nextLineNumber);
        while (nextLine && nextLine.search(MarkdownEditorBase.ORDERED_LIST_PATTERN) !== -1) {
            var listNumberString = "" + ++listNumber;
            var dotPos = nextLine.search(/\./);
            this.cm.replaceRange(listNumberString, { line: nextLineNumber, ch: 0 }, { line: nextLineNumber, ch: dotPos }, '+replaceTokenAtLineStart');
            nextLine = this.cm.getLine(++nextLineNumber);
        }
    };
    /**
     * Toggle "check list" for each selected line. Furthermore, a selected (un)ordered list line is
     * transformed to a check list.
     */
    MarkdownEditorBase.prototype.toggleCheckList = function () {
        var preferred = this.options.preferredTokens.checkList + ' [ ] ';
        this.replaceTokenAtLineStart(function (oldLineContent) {
            // Has selected line a bullet point token?
            if (oldLineContent.search(MarkdownEditorBase.CHECK_LIST_PATTERN) === -1) {
                // Has selected line an enumeration token?
                if (oldLineContent.search(MarkdownEditorBase.ORDERED_LIST_PATTERN) !== -1) {
                    return oldLineContent.replace(MarkdownEditorBase.ORDERED_LIST_PATTERN, preferred);
                }
                // Has selected line a check list token?
                if (oldLineContent.search(MarkdownEditorBase.UNORDERED_LIST_PATTERN) !== -1) {
                    return oldLineContent.replace(MarkdownEditorBase.UNORDERED_LIST_PATTERN, preferred);
                }
                return preferred + oldLineContent;
            }
            else {
                return oldLineContent.replace(MarkdownEditorBase.CHECK_LIST_PATTERN, '');
            }
        });
    };
    /**
     * Replace each selected line with the result of the callback function `replaceFn`.
     * Additionally adjusts the selection boundaries to the originally selected boundaries.
     * @param replaceFn callback function to the calculate the line replacements
     */
    MarkdownEditorBase.prototype.replaceTokenAtLineStart = function (replaceFn) {
        var newSelections = [];
        for (var _i = 0, _a = this.cm.listSelections(); _i < _a.length; _i++) {
            var sel = _a[_i];
            var selection = _.cloneDeep(sel);
            var shiftFrom = 0;
            var shiftTo = 0;
            for (var lineNumber = selection.from().line; lineNumber <= selection.to().line; lineNumber++) {
                var oldLineContent = this.cm.getLine(lineNumber);
                var newLineContent = replaceFn(oldLineContent, lineNumber);
                this.cm.replaceRange(newLineContent, { line: lineNumber, ch: 0 }, { line: lineNumber, ch: oldLineContent.length }, '+replaceTokenAtLineStart');
                // Set shifts for selection start and end
                if (lineNumber === selection.from().line) {
                    shiftFrom = newLineContent.length - oldLineContent.length;
                }
                if (lineNumber === selection.to().line) {
                    shiftTo = newLineContent.length - oldLineContent.length;
                }
            }
            // Adjust selection boundaries to originally selected boundaries
            if (_.isEqual(selection.anchor, selection.from())) {
                selection.anchor.ch += shiftFrom;
                if (!selection.empty())
                    selection.head.ch += shiftTo;
            }
            else {
                selection.anchor.ch += shiftTo;
                if (!selection.empty())
                    selection.head.ch += shiftFrom;
            }
            newSelections.push(selection);
        }
        this.cm.setSelections(newSelections, undefined, { origin: 'replaceTokenAtLineStart' });
        this.cm.focus();
    };
    /**
     * Wrap each selection with code block tokens, which are inserted in separate lines.
     */
    MarkdownEditorBase.prototype.insertCodeBlock = function () {
        var newSelections = [];
        var selections = _.cloneDeep(this.cm.listSelections());
        for (var i = 0; i < selections.length; i++) {
            var oldSelection = selections[i];
            var newSelection = _.cloneDeep(oldSelection);
            var preferredToken = this.options.preferredTokens.codeBlock;
            // Wrap selection with code block tokens
            var currentShift = 3;
            var startToken = preferredToken + '\n';
            if (newSelection.from().ch > 0) {
                startToken = '\n' + startToken;
                currentShift++;
            }
            this.cm.replaceRange('\n' + preferredToken + '\n', newSelection.to(), undefined, '+insertCodeBlock');
            this.cm.replaceRange(startToken, newSelection.from(), undefined, '+insertCodeBlock');
            // Adjust selections to originally selected characters
            var offset = newSelection.from().line === newSelection.to().line ? newSelection.from().ch : 0;
            if (!newSelection.empty()) {
                newSelection.to().line += currentShift - 2;
                newSelection.to().ch -= offset;
            }
            else {
                // fix for edge case bug of empty selection with not synchronous anchor and head
                newSelection.anchor = newSelection.head;
            }
            newSelection.from().line += currentShift - 2;
            newSelection.from().ch = 0;
            newSelections.push(newSelection);
            // Adjust all following selections to originally selected characters
            for (var j = i + 1; j < selections.length; j++) {
                var s = selections[j];
                var remainderIndex = oldSelection.to().ch;
                if (s.from().line === oldSelection.to().line) {
                    s.from().ch -= remainderIndex;
                }
                if (s.to().line === oldSelection.to().line) {
                    s.to().ch -= remainderIndex;
                }
                if (!s.empty())
                    s.to().line += currentShift;
                s.from().line += currentShift;
            }
        }
        this.cm.setSelections(newSelections, undefined, { origin: '+insertCodeBlock' });
        this.cm.focus();
    };
    /**
     * Wrap a link template around each selection.
     */
    MarkdownEditorBase.prototype.insertLink = function () {
        var _a = this.options.preferredTemplates.link, before = _a[0], after = _a[1];
        this.insertInlineTemplate(before, after);
    };
    /**
     * Wrap a image link template around each selection.
     */
    MarkdownEditorBase.prototype.insertImageLink = function () {
        var _a = this.options.preferredTemplates.imageLink, before = _a[0], after = _a[1];
        this.insertInlineTemplate(before, after);
    };
    /**
     * Wrap a template around each selection with the specified `before` and `after` template parts.
     * @param before the template part inserted **before** the selection start
     * @param after the template part inserted **after** the selection start
     */
    MarkdownEditorBase.prototype.insertInlineTemplate = function (before, after) {
        var newSelections = [];
        var selections = this.cm.listSelections();
        for (var i = 0; i < selections.length; i++) {
            var oldSelection = selections[i];
            var newSelection = _.cloneDeep(oldSelection);
            // Insert template parts before and after the selection
            this.cm.replaceRange(after, newSelection.to(), undefined, '+toggleBlock');
            this.cm.replaceRange(before, newSelection.from(), undefined, '+toggleBlock');
            // Adjust selections to originally selected characters
            if (newSelection.empty())
                newSelection.head = newSelection.anchor;
            else
                newSelection.to().ch += before.length;
            newSelection.from().ch += before.length;
            newSelections.push(newSelection);
            // Adjust all following selections to originally selected characters
            for (var j = i + 1; j < selections.length; j++) {
                var s = selections[j];
                if (s.empty()) {
                    s.head = s.anchor;
                }
                else {
                    if (s.head.line === oldSelection.from().line)
                        s.head.ch += before.length;
                    if (s.head.line === oldSelection.to().line)
                        s.head.ch += after.length;
                }
                if (s.anchor.line === oldSelection.from().line)
                    s.anchor.ch += before.length;
                if (s.anchor.line === oldSelection.to().line)
                    s.anchor.ch += after.length;
                else
                    break;
            }
        }
        this.cm.setSelections(newSelections, undefined, { origin: '+toggleBlock' });
        this.cm.focus();
    };
    /**
     * Insert a horizontal rule in the subsequent line of each selection.
     */
    MarkdownEditorBase.prototype.insertHorizontalRule = function () {
        var preferred = this.options.preferredTokens.horizontalRule;
        this.insertBlockTemplateBelow("\n" + preferred + "\n\n");
    };
    /**
     * Insert a table template with the specified number of rows and columns
     * in the subsequent line of each selection.
     * @param rows number of rows
     * @param columns number columns
     */
    MarkdownEditorBase.prototype.insertTable = function (rows, columns) {
        var tableOptions = this.options.preferredTemplates.table;
        if (typeof tableOptions === 'string') {
            this.insertBlockTemplateBelow(tableOptions);
        }
        else {
            if (!rows) {
                rows = tableOptions.rows;
            }
            if (!columns) {
                columns = tableOptions.columns;
            }
            var template = '\n';
            for (var c = 1; c <= columns; c++) {
                template += "| Column " + c + " ";
            }
            template += '|\n';
            for (var c = 1; c <= columns; c++) {
                template += '| -------- ';
            }
            template += '|\n';
            for (var r = 1; r <= rows; r++) {
                for (var c = 1; c <= columns; c++) {
                    template += '| Content  ';
                }
                template += '|\n';
            }
            template += '\n';
            this.insertBlockTemplateBelow(template);
        }
    };
    /**
     * Insert a block template in the subsequent line of each selection.
     * The template can contain multiple lines separated with the specified `lineSeparator`.
     * @param template the template
     * @param lineSeparator The line separator. Default is `\n`.
     */
    MarkdownEditorBase.prototype.insertBlockTemplateBelow = function (template, lineSeparator) {
        var _a;
        if (lineSeparator === void 0) { lineSeparator = '\n'; }
        if (lineSeparator !== '\n')
            template = template.replace(RegExp(lineSeparator, 'g'), '\n');
        var currentShift = 0; // indicates how many lines have been inserted
        var selections = this.cm.listSelections();
        for (var i = 0; i < selections.length; i++) {
            var oldSelection = selections[i];
            var newSelection = _.cloneDeep(oldSelection);
            // Shift selection back to correct position in respect to previously inserted lines
            if (newSelection.empty())
                newSelection.head = newSelection.anchor;
            else
                newSelection.to().line += currentShift;
            newSelection.from().line += currentShift;
            var toLineNumber = newSelection.to().line;
            var toLineLength = this.cm.getLine(toLineNumber).length;
            var toLineEnd = { line: toLineNumber, ch: toLineLength };
            if (toLineLength > 0) {
                template = '\n' + template;
            }
            currentShift += ((_a = template.match(/\n/g)) === null || _a === void 0 ? void 0 : _a.length) || 0;
            // Insert template in the subsequent line of the selection
            this.cm.replaceRange(template, toLineEnd, undefined, '+insertHorizontalRule');
        }
        this.cm.focus();
    };
    /**
     * Open a Markdown Guide in a new tab.
     */
    MarkdownEditorBase.prototype.openMarkdownGuide = function () {
        window.open(this.options.markdownGuideUrl, '_blank');
    };
    /***** Extended Editor API *****/
    /**
     * Undo one edit. Shortcut for `Codemirror.undo()`.
     */
    MarkdownEditorBase.prototype.undo = function () {
        this.cm.undo();
    };
    /**
     * Redo one edit. Shortcut for `Codemirror.redo()`.
     */
    MarkdownEditorBase.prototype.redo = function () {
        this.cm.redo();
    };
    /**
     * Toggle the editor's rich-text mode. If off, there is no markdown styling inside the editor.
     */
    MarkdownEditorBase.prototype.toggleRichTextMode = function () {
        var currentMode = this.cm.getOption('mode');
        if (currentMode === 'gfm' || currentMode.name === 'gfm') {
            this.cm.setOption('mode', '');
        }
        else {
            this.cm.setOption('mode', this.getGfmMode());
        }
    };
    /**
     * Start a file download containing the editor content as plain text.
     * The name of the file is either `fileName` if specified, or the default `downloadFileName`
     * specified in the Markdown Editor options.
     * @param fileName The name of the downloaded file. Preferred over `options.downloadFileName`.
     */
    MarkdownEditorBase.prototype.downloadAsFile = function (fileName) {
        var data = new Blob([this.getContent()], { type: 'text/plain' });
        var url = window.URL.createObjectURL(data);
        var a = document.createElement('a');
        a.setAttribute('href', url);
        a.setAttribute('download', fileName || this.options.downloadFileNameGenerator());
        a.click();
    };
    /**
     * Import content from a file. Reads the content of `file` if specified, or opens the browser's
     * file selection dialog and reads the content of the user-selected file.
     * @param file The file to import the content from.
     */
    MarkdownEditorBase.prototype.importFromFile = function (file) {
        var _this = this;
        var readFile = function (file) {
            var reader = new FileReader();
            reader.onload = function (event) { return _this.setContent((event.target.result || '')); };
            reader.readAsText(file);
        };
        if (file) {
            readFile(file);
        }
        else {
            var input = document.createElement('input');
            input.setAttribute('type', 'file');
            input.setAttribute('accept', '.txt, .text, .markdown, .mdown, .mkdn, .md, .mkd, .mdwn, .mdtxt, .mdtext');
            input.onchange = function (event) {
                var files = (event === null || event === void 0 ? void 0 : event.target).files;
                if (files) {
                    readFile(files[0]);
                }
            };
            input.click();
        }
    };
    /**
     * Formats the content using Prettier's Markdown parser.
     */
    MarkdownEditorBase.prototype.formatContent = function () {
        this.setContent(prettier.format(this.getContent(), {
            parser: 'markdown',
            plugins: [parserMarkdown],
        }));
    };
    /***** Developer API *****/
    /**
     * Get the editor's content with the specified line break format.
     * @param lineSeparator The line break format. Default is `\n`.
     */
    MarkdownEditorBase.prototype.getContent = function (lineSeparator) {
        if (lineSeparator === void 0) { lineSeparator = '\n'; }
        return this.cm.getValue(lineSeparator);
    };
    /**
     * Get the editor's content with the lines as array.
     */
    MarkdownEditorBase.prototype.getContentPerLine = function () {
        return this.cm.getValue().split('\n');
    };
    /**
     * Set the editor's content with the specified line break format.
     * @param content The content.
     * @param lineSeparator The line break format. Default is `\n`.
     */
    MarkdownEditorBase.prototype.setContent = function (content, lineSeparator) {
        if (lineSeparator === void 0) { lineSeparator = '\n'; }
        if (lineSeparator !== '\n')
            content = content.replace(RegExp(lineSeparator, 'g'), '\n');
        this.cm.setValue(content);
    };
    /**
     * Get the editor's content with the lines as array.
     */
    MarkdownEditorBase.prototype.setContentPerLine = function (content) {
        this.cm.setValue(content.join('\n'));
    };
    /**
     * Focus the editor. Shortcut for `Codemirror.focus()`.
     */
    MarkdownEditorBase.prototype.focus = function () {
        this.cm.focus();
    };
    /**
     * Get the number of characters in the document.
     */
    MarkdownEditorBase.prototype.getCharacterCount = function () {
        // eslint-disable-next-line no-control-regex
        return this.cm.getValue().replace(RegExp('\n', 'gi'), '').length;
    };
    /**
     * Get the number of words in the document.
     */
    MarkdownEditorBase.prototype.getWordCount = function () {
        var content = this.cm.getValue();
        var s = content.replace(/(^\s*)|(\s*$)/gi, ''); // Trim left and right
        s = s.replace(/\t/gi, ' '); // Replace tabs by single whitespace
        s = s.replace(/\n/gi, ' '); // Replace line breaks by single whitespace
        s = s.replace(/[ ]{2,}/gi, ' '); // Reduce multiple whitespaces to a single one
        var split = s.split(' '); // Split string by whitespaces
        split = split.filter(function (x) { return !/^\W+$/.test(x); }); // Filter out non-alphanumeric words
        return s === '' ? 0 : split.length;
    };
    /**
     * Get the current cursor position as a `{line, ch}` object.
     * Shortcut for `Codemirror.getCursor()`.
     */
    MarkdownEditorBase.prototype.getCursorPos = function () {
        return this.cm.getCursor();
    };
    /**
     * Returns whether the document has been modified.
     * Inverted shortcut for `Codemirror.isClean()`
     */
    MarkdownEditorBase.prototype.isDirty = function () {
        return !this.cm.isClean();
    };
    /**
     * Check if text at current cursor position has specified token type
     * like returned by `CodeMirror.getTokenTypeAt()`.
     * @param token
     */
    MarkdownEditorBase.prototype.hasTokenAtCursorPos = function (token) {
        var _a;
        return ((_a = this.cm.getTokenTypeAt(this.getCursorPos())) === null || _a === void 0 ? void 0 : _a.split(' ').includes(token)) || false;
    };
    /**
     * Get the type of list in the specified line. Return undefined, if the line contains no list token.
     * @param lineNumber
     */
    MarkdownEditorBase.prototype.getListTypeOfLine = function (lineNumber) {
        var line = this.cm.getLine(lineNumber);
        if (MarkdownEditorBase.CHECK_LIST_PATTERN.test(line))
            return 'check';
        if (MarkdownEditorBase.UNORDERED_LIST_PATTERN.test(line))
            return 'unordered';
        if (MarkdownEditorBase.ORDERED_LIST_PATTERN.test(line))
            return 'ordered';
        return undefined;
    };
    /**
     * Get the action shortcuts effectively applied to the editor.
     *
     * This might be a mix of custom shortcuts
     * and default shortcuts as specified in `DEFAULT_OPTIONS`.
     */
    MarkdownEditorBase.prototype.getShortcuts = function () {
        return this.options.shortcuts;
    };
    /**
     * Add the specified keyboard shortcut to the specified action function.
     *
     * **Note:** This does **not** replace an existing shortcut to the specified action,
     * but only **adds** it. In case of an already existing shortcut, the result will be
     * two distinct shortcuts to the same function. However, you can use `removeShortcut()`
     * to remove the old one.
     * @param hotkeys the hotkeys string as required by Codemirror
     * @param action the action function
     * @see https://codemirror.net/doc/manual.html#keymaps
     */
    MarkdownEditorBase.prototype.addShortcut = function (hotkeys, action) {
        var extraKeys = (this.cm.getOption('extraKeys') || {});
        var shortcut;
        if (isMac()) {
            shortcut = hotkeys.replace('Ctrl', 'Cmd');
        }
        else {
            shortcut = hotkeys.replace('Cmd', 'Ctrl');
        }
        extraKeys[shortcut] = action;
        this.cm.setOption('extraKeys', extraKeys);
    };
    /**
     * Remove the specified keyboard shortcut.
     *
     * @param hotkeys the hotkeys string as required by Codemirror
     * @see https://codemirror.net/doc/manual.html#keymaps
     */
    MarkdownEditorBase.prototype.removeShortcut = function (hotkeys) {
        var extraKeys = (this.cm.getOption('extraKeys') || {});
        var shortcut;
        if (isMac()) {
            shortcut = hotkeys.replace('Ctrl', 'Cmd');
        }
        else {
            shortcut = hotkeys.replace('Cmd', 'Ctrl');
        }
        delete extraKeys[shortcut];
        this.cm.setOption('extraKeys', extraKeys);
    };
    /***** Markdown Editor Options *****/
    /**
     * Overwrites current options with specified options.
     * Options that are not included in specified `options` will not be modified.
     * @param options the set of options that shall be changed
     */
    MarkdownEditorBase.prototype.setOptions = function (options) {
        if (!options)
            return;
        this.options = _.merge(this.options, options);
        this.applyCodemirrorOptions();
        this.applyEditorKeyMappings();
    };
    /**
     * Apply codemirror-specific options that are specified in `this.options`.
     */
    MarkdownEditorBase.prototype.applyCodemirrorOptions = function () {
        var _a;
        this.cm.setOption('autofocus', this.options.autofocus);
        this.cm.setOption('configureMouse', this.options.multipleCursors ? undefined : function () { return ({ addNew: false }); });
        this.cm.setOption('lineNumbers', this.options.lineNumbers);
        this.cm.setOption('lineWrapping', this.options.lineWrapping);
        this.cm.setOption('placeholder', this.options.placeholder);
        this.cm.setOption('readOnly', this.options.disabled);
        this.cm.setOption('tabSize', this.options.tabSize);
        this.cm.setOption('theme', this.options.themes.join(' '));
        if (this.options.richTextMode) {
            if (((_a = this.cm.getOption('mode')) === null || _a === void 0 ? void 0 : _a.name) !== 'gfm') {
                this.cm.setOption('mode', this.getGfmMode());
            }
        }
        else {
            this.cm.setOption('mode', '');
        }
    };
    /**
     * Apply the key map for markdown editor actions specified in `this.options.shortcuts`
     * to the Codemirror editor instance.
     */
    MarkdownEditorBase.prototype.applyEditorKeyMappings = function () {
        var _this = this;
        if (this.options.shortcutsEnabled === 'none') {
            this.cm.setOption('extraKeys', {});
            return;
        }
        var bindings = {
            increaseHeadingLevel: function () { return _this.increaseHeadingLevel(); },
            decreaseHeadingLevel: function () { return _this.decreaseHeadingLevel(); },
            toggleBold: function () { return _this.toggleBold(); },
            toggleItalic: function () { return _this.toggleItalic(); },
            toggleStrikethrough: function () { return _this.toggleStrikethrough(); },
            toggleUnorderedList: function () { return _this.toggleUnorderedList(); },
            toggleOrderedList: function () { return _this.toggleOrderedList(); },
            toggleCheckList: function () { return _this.toggleCheckList(); },
            toggleQuote: function () { return _this.toggleQuote(); },
            insertLink: function () { return _this.insertLink(); },
            insertImageLink: function () { return _this.insertImageLink(); },
            insertTable: function () { return _this.insertTable(); },
            insertHorizontalRule: function () { return _this.insertHorizontalRule(); },
            toggleInlineCode: function () { return _this.toggleInlineCode(); },
            insertCodeBlock: function () { return _this.insertCodeBlock(); },
            openMarkdownGuide: function () { return _this.openMarkdownGuide(); },
            toggleRichTextMode: function () { return _this.toggleRichTextMode(); },
            downloadAsFile: function () { return _this.downloadAsFile(); },
            importFromFile: function () { return _this.importFromFile(); },
            formatContent: function () { return _this.formatContent(); },
        };
        var shortcuts = this.options.shortcuts;
        var extraKeys = {};
        for (var _i = 0, _a = Object.entries(shortcuts); _i < _a.length; _i++) {
            var _b = _a[_i], key = _b[0], value = _b[1];
            if (value) {
                var shortcut = void 0;
                if (isMac()) {
                    shortcut = value.replace('Ctrl', 'Cmd');
                }
                else {
                    shortcut = value.replace('Cmd', 'Ctrl');
                }
                if (extraKeys[shortcut]) {
                    console.warn("Caution: Duplicate keybinding! Shortcut '" + shortcut + "' was already " +
                        ("bound to action '" + extraKeys[shortcut].toString().slice(27, -5) + "' ") +
                        ("and is now overridden with action '" + key + "'."));
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any
                }
                else if (CodeMirror.keyMap.default[shortcut]) {
                    console.warn("Caution: Duplicate keybinding! Shortcut '" + shortcut + "' was already " +
                        (
                        // eslint-disable-next-line @typescript-eslint/no-explicit-any
                        "bound to native CodeMirror action '" + CodeMirror.keyMap.default[shortcut] + "' ") +
                        ("and is now overridden with action '" + key + "'."));
                }
                extraKeys[shortcut] = bindings[key];
            }
        }
        // if (!(_.matches(this.cm.getOption('extraKeys'))(extraKeys)))
        this.cm.setOption('extraKeys', extraKeys);
    };
    /***** Private methods *****/
    /**
     * Create the GFM mode object. It includes some adjustments for the default
     * theme to use Codemirror's default markup styling while always providing
     * expressive class names for the tokens.
     */
    MarkdownEditorBase.prototype.getGfmMode = function () {
        var isDefaultTheme = this.options.themes === undefined || this.options.themes.includes('default');
        return {
            name: 'gfm',
            tokenTypeOverrides: {
                header: 'header',
                code: 'code' + (isDefaultTheme ? ' comment' : ''),
                quote: 'quote',
                list1: 'list list-level-1' + (isDefaultTheme ? ' variable-2' : ''),
                list2: 'list list-level-2' + (isDefaultTheme ? ' variable-3' : ''),
                list3: 'list list-level-gt-2' + (isDefaultTheme ? ' keyword' : ''),
                hr: 'hr',
                image: 'image',
                imageAltText: 'image-alt-text' + (isDefaultTheme ? ' string' : ''),
                imageMarker: 'image-marker',
                formatting: 'token',
                linkInline: 'link link-inline',
                linkEmail: 'link link-email',
                linkText: 'link-text' + (isDefaultTheme ? ' string' : ''),
                linkHref: 'link link-href',
                em: 'em',
                strong: 'strong',
                strikethrough: 'strikethrough',
                emoji: 'emoji' + (isDefaultTheme ? ' builtin' : ''),
            },
            highlightFormatting: this.options.highlightTokens,
        };
    };
    /**
     * Remove the `.cm-link` class from images' alt texts because text is not a link.
     * This has been hardcoded to Codemirror's Markdown mode and thus cannot be changed
     * without compiling Codemirror ourselves within this project, which is not the
     * desired way so far.
     */
    MarkdownEditorBase.prototype.removeLinkClassFromImageTexts = function () {
        if (!this.getGfmMode().tokenTypeOverrides.imageAltText.includes('link')) {
            this.cm.on('changes', function () {
                var codemirrorEl = document.querySelector('.CodeMirror');
                codemirrorEl.querySelectorAll('.cm-image-alt-text').forEach(function (el) { return el.classList.remove('cm-link'); });
            });
        }
    };
    MarkdownEditorBase.ORDERED_LIST_PATTERN = /^( )*(\d)+\.(\t| )+/;
    MarkdownEditorBase.UNORDERED_LIST_PATTERN = /^( )*(\*|-)(\t| )+/;
    MarkdownEditorBase.CHECK_LIST_PATTERN = /^( )*(\*|-) \[(X|x| )\](\t| )+/;
    MarkdownEditorBase.BOLD_TOKENS = ['**', '__'];
    MarkdownEditorBase.ITALIC_TOKENS = ['*', '_'];
    return MarkdownEditorBase;
}());
var MarkdownEditor = /** @class */ (function (_super) {
    __extends(MarkdownEditor, _super);
    function MarkdownEditor(element, options) {
        var _this = this;
        var opts = _.merge(_.cloneDeep(DEFAULT_OPTIONS), options);
        if (opts.shortcutsEnabled === 'customOnly') {
            opts.shortcuts = (options === null || options === void 0 ? void 0 : options.shortcuts) || {};
        }
        var cm = CodeMirror(element, opts);
        _this = _super.call(this, cm, opts) || this;
        return _this;
    }
    return MarkdownEditor;
}(MarkdownEditorBase));
export { MarkdownEditor };
var MarkdownEditorFromTextarea = /** @class */ (function (_super) {
    __extends(MarkdownEditorFromTextarea, _super);
    function MarkdownEditorFromTextarea(textarea, options) {
        var _this = this;
        var opts = _.merge(_.cloneDeep(DEFAULT_FROM_TEXTAREA_OPTIONS), options);
        if (opts.shortcutsEnabled === 'customOnly') {
            opts.shortcuts = (options === null || options === void 0 ? void 0 : options.shortcuts) || {};
        }
        var cm = CodeMirror.fromTextArea(textarea);
        _this = _super.call(this, cm, opts) || this;
        _this.cm = cm;
        _this.options = opts;
        return _this;
    }
    /**
     * Copy the content of the editor into the textarea.
     *
     * Shortcut for `Codemirror.EditorFromTextarea.save()`.
     */
    MarkdownEditorFromTextarea.prototype.syncTextarea = function () {
        this.cm.save();
    };
    /**
     * Remove the editor, and restore the original textarea (with the editor's current content).
     *
     * Shortcut for `Codemirror.EditorFromTextarea.toTextarea()`.
     */
    MarkdownEditorFromTextarea.prototype.toTextarea = function () {
        this.cm.toTextArea();
    };
    /**
     * Returns the textarea that the instance was based on.
     *
     * Shortcut for `Codemirror.EditorFromTextarea.getTextarea()`.
     */
    MarkdownEditorFromTextarea.prototype.getTextarea = function () {
        return this.cm.getTextArea();
    };
    /**
     * @inheritdoc
     */
    MarkdownEditorFromTextarea.prototype.setOptions = function (options) {
        var _this = this;
        if (!options)
            return;
        var opts = _.merge(this.options, options);
        _super.prototype.setOptions.call(this, opts);
        this.options = opts;
        if (this.saver === undefined && this.options.autoSync) {
            this.saver = function () { return _this.syncTextarea(); };
            this.cm.on('changes', this.saver);
        }
        else if (this.saver !== undefined && !this.options.autoSync) {
            this.cm.off('changes', this.saver);
            this.saver = undefined;
        }
    };
    return MarkdownEditorFromTextarea;
}(MarkdownEditorBase));
export { MarkdownEditorFromTextarea };
/***** Util ******/
/**
 * Transform a string so that all Regex-reserved characters are escaped.
 * @param s
 */
function escapeRegexChars(s) {
    var reservedChars = ['.', '*', '+', '?', '!', '{', '}', '[', ']'];
    s = s.replace(/\\/gi, '\\\\');
    for (var _i = 0, reservedChars_1 = reservedChars; _i < reservedChars_1.length; _i++) {
        var reservedChar = reservedChars_1[_i];
        s = s.replace(RegExp("\\" + reservedChar, 'gi'), '\\' + reservedChar);
    }
    return s;
}
function isMac() {
    return /Mac/.test(navigator.platform);
}
//# sourceMappingURL=markdown-editor.js.map